/*

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "quakedef.h"
#include "snd_main.h"

#include <CoreAudio/AudioHardware.h>

/*

CoreAudio support for OS X - BigMac

*/

#define OUTPUT_BUFFER_SIZE	(4 * 1024)
#define TOTAL_BUFFER_SIZE	(64 * 1024)

int	samplebits;		// can be 8 (low quality) or 16 (high quality)
int	submission_chunk;

static AudioDeviceID				gSNDDMASoundDeviceID;
static AudioStreamBasicDescription	gSNDDMABasicDescription;

static unsigned char	gSNDDMABuffer[TOTAL_BUFFER_SIZE];
static UInt32			gSNDDMABufferPosition, gSNDDMABufferByteCount;

static OSStatus	(*SNDDMA_AudioIOProc)(AudioDeviceID, const AudioTimeStamp *,
										const AudioBufferList *, const AudioTimeStamp *,
										AudioBufferList *, const AudioTimeStamp *, void *);

/*
static OSStatus SNDDMA_Audio8BitIOProc (AudioDeviceID inDevice,
										const AudioTimeStamp *inNow,
										const AudioBufferList *inInputData,
										const AudioTimeStamp *inInputTime,
										AudioBufferList *outOutputData, 
										const AudioTimeStamp *inOutputTime,
										void *inClientData)
{
    UInt16	i;
    UInt8	*myDMA			= ((UInt8 *) gSNDDMABuffer) + gSNDDMABufferPosition / (samplebits >> 3);
    float	*myOutBuffer	= (float *) outOutputData->mBuffers[0].mData;
	float	bufScale		= (1.0f / 128.0f);

    // convert the buffer from unsigned PCM to signed PCM and then to float, required by CoreAudio:
    for (i=0; i < gSNDDMABufferByteCount; i++)
        *myOutBuffer++ = ((*myDMA++) - 0x80) * bufScale;
    
    // increase the bufferposition:
    gSNDDMABufferPosition += gSNDDMABufferByteCount * (samplebits >> 3);
    if (gSNDDMABufferPosition >= sizeof (gSNDDMABuffer))
        gSNDDMABufferPosition = 0;
    
    // return 0 = no error:
    return 0;
}
*/

static OSStatus SNDDMA_Audio16BitIOProc (AudioDeviceID inDevice,
                                         const AudioTimeStamp *inNow,
                                         const AudioBufferList *inInputData,
                                         const AudioTimeStamp *inInputTime,
                                         AudioBufferList *outOutputData, 
                                         const AudioTimeStamp *inOutputTime,
                                         void *inClientData)
{
    UInt16	i;
    short	*myDMA			= ((short *) gSNDDMABuffer) + gSNDDMABufferPosition / (samplebits >> 3);
    float	*myOutBuffer	= (float *) outOutputData->mBuffers[0].mData;
	float	bufScale		= (1.0f / 32768.0f);

    // convert the buffer to float, required by CoreAudio:
    for (i=0; i < gSNDDMABufferByteCount; i++)
        *myOutBuffer++ = (*myDMA++) * bufScale;

    // increase the bufferposition:
    gSNDDMABufferPosition += gSNDDMABufferByteCount * (samplebits >> 3);
    if (gSNDDMABufferPosition >= sizeof (gSNDDMABuffer))
        gSNDDMABufferPosition = 0;

    // return 0 = no error:
    return 0;
}

qboolean SNDDMA_ReserveBufferSize (void)
{
    OSStatus		myError;
    AudioDeviceID	myAudioDevice;
    UInt32			myPropertySize;

    myPropertySize = sizeof (AudioDeviceID);
    myError = AudioHardwareGetProperty (kAudioHardwarePropertyDefaultOutputDevice, &myPropertySize,&myAudioDevice);

    if (!myError && myAudioDevice != kAudioDeviceUnknown)
    {
        UInt32	myBufferByteCount = OUTPUT_BUFFER_SIZE * sizeof (float);

        myPropertySize = sizeof (myBufferByteCount);

        // set the buffersize for the audio device:
        myError = AudioDeviceSetProperty (myAudioDevice, NULL, 0, false, kAudioDevicePropertyBufferSize, myPropertySize, &myBufferByteCount);

        if (!myError)
            return true;
    }

    return false;
}

/*
==================
SNDDMA_Init

Try to find a sound device to mix for.
Returns false if nothing is found.
==================
*/
qboolean SNDDMA_Init(void)
{
	UInt32	myPropertySize;
	
	if (! SNDDMA_ReserveBufferSize () )
		Con_Printf ("CoreAudio: ReserveBufferSize failed.\n");

	samplebits = 16;	// quality (this should be configurable)

	//if (samplebits == 8)
	//	SNDDMA_AudioIOProc = SNDDMA_Audio8BitIOProc;
	//else
		SNDDMA_AudioIOProc = SNDDMA_Audio16BitIOProc;

    myPropertySize = sizeof (gSNDDMASoundDeviceID);

    // find a suitable audio device:
    if (AudioHardwareGetProperty (kAudioHardwarePropertyDefaultOutputDevice, &myPropertySize, &gSNDDMASoundDeviceID))
    {
        Con_Printf ("CoreAudio: Can\'t get audio device.\n");
        return false;
    }
	
    // is the device valid?
    if (gSNDDMASoundDeviceID == kAudioDeviceUnknown)
    {
        Con_Printf ("CoreAudio: Unsupported audio device.\n");
        return false;
    }
	
    // get the buffersize of the audio device [must previously be set via "SNDDMA_ReserveBufferSize ()"]:
    myPropertySize = sizeof (gSNDDMABufferByteCount);
    if (AudioDeviceGetProperty (gSNDDMASoundDeviceID, 0, false, kAudioDevicePropertyBufferSize,
                                &myPropertySize, &gSNDDMABufferByteCount) || gSNDDMABufferByteCount == 0)
    {
        Con_Printf ("CoreAudio: Can't get audiobuffer.\n");
        return false;
    }
	
    //check the buffersize:
    gSNDDMABufferByteCount /= sizeof (float);

    if (gSNDDMABufferByteCount != OUTPUT_BUFFER_SIZE)
        Con_Printf ("CoreAudio: Audiobuffer size is not sufficient for clean movie playback!\n");

    if (sizeof (gSNDDMABuffer) % gSNDDMABufferByteCount != 0 || sizeof (gSNDDMABuffer) / gSNDDMABufferByteCount < 2)
    {
        Con_Printf ("CoreAudio: Bad audiobuffer size!\n");
        return false;
    }
	
	// get the audiostream format:
    myPropertySize = sizeof (gSNDDMABasicDescription);
    if (AudioDeviceGetProperty (gSNDDMASoundDeviceID, 0, false, kAudioDevicePropertyStreamFormat,
                                &myPropertySize, &gSNDDMABasicDescription))
    {
        Con_Printf ("CoreAudio init failed.\n");
        return false;
    }
	
    // is the format LinearPCM?
    if (gSNDDMABasicDescription.mFormatID != kAudioFormatLinearPCM)
    {
        Con_Printf ("Default Audio Device doesn't support Linear PCM!\n");
        return false;
    }
	
	// add the sound FX IO:
	if (AudioDeviceAddIOProc (gSNDDMASoundDeviceID, SNDDMA_AudioIOProc, NULL))
	{
		Con_Printf ("CoreAudio: Can\'t install IOProc.\n");
		return false;
	}
        
	// start the sound FX:
	if (AudioDeviceStart (gSNDDMASoundDeviceID, SNDDMA_AudioIOProc))
	{
		Con_Printf ("CoreAudio: Can\'t start audio.\n");
		return false;
	}

	// Init the shm structure
	memset((void*) shm, 0, sizeof(*shm));

	shm->format.channels = gSNDDMABasicDescription.mChannelsPerFrame;
	shm->format.width = samplebits / 8;	// i think this is right
	shm->format.speed = gSNDDMABasicDescription.mSampleRate;

	shm->samples = sizeof (gSNDDMABuffer) / (samplebits >> 3);
	shm->samplepos = 0;

	shm->bufferlength = shm->samples * shm->format.width;
	shm->buffer = gSNDDMABuffer;

	submission_chunk = gSNDDMABufferByteCount;

	gSNDDMABufferPosition = 0;

	Con_Printf ("CoreAudio initialized (%i channels, %i bits).\n", shm->format.channels, samplebits);
	return true;
}

/*
==============
SNDDMA_GetDMAPos

return the current sample position (in mono samples read)
inside the recirculating dma buffer, so the mixing code will know
how many sample are required to fill it up.
===============
*/
int SNDDMA_GetDMAPos(void)
{
	return (gSNDDMABufferPosition / (samplebits >> 3));
}

/*
==============
SNDDMA_Shutdown

Reset the sound device for exiting
===============
*/
void SNDDMA_Shutdown(void)
{
	AudioDeviceStop (gSNDDMASoundDeviceID, SNDDMA_AudioIOProc);
	AudioDeviceRemoveIOProc (gSNDDMASoundDeviceID, SNDDMA_AudioIOProc);
}

void SNDDMA_Submit(void)
{
	// not required
}

void *S_LockBuffer(void)
{
	// not required (just return the buffer)
	return shm->buffer;
}


void S_UnlockBuffer(void)
{
	// not required
}